/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"

//`define S0_INCLUDE_ROM

// Module Name: rtavr_s0_fetch
//   Ver 005
//
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                               NORMAL     w 8K ROM
// Number of Slice Flip Flops:        4         25  
// Number of 4 input LUTs:           26         36 
// Number of occupied Slices:        16         32    
// Total Number of 4 input LUTs:     26         36   
// Number of bonded IOBs :           54         81  
//   IOB Flip Flops :                16          -    
// Number of BUFGMUXs :               1          1   
// Number of RAMB16BWEs :             -          4  
// Clock to Setup 
//on destination clock CLK            ?      3.809|    

module rtavr_s0_fetch (
		input CLK,
		input RESET,
`ifdef S0_INCLUDE_ROM
		input [11:0] PC,
		input WEB, OE, 
		input [12:0] ADDRB,
		input [7:0] DIB, 
		output[7:0] DOB,
`else
		input [15:0] PM_OUT,
`endif
		input S0_VALID,
		output WB_INST,
`ifdef NO_STALL_IDX
`else
		output IDX_INST,
`endif
		output [15:0] INST,
		output GPR_PREDEC, GPR_POSTINC,
		output[3:0] GPR_ADDRAL,
		output[3:0] GPR_ADDRAH,
		output[3:0] GPR_ADDRBL,
		output[3:0] GPR_ADDRBH
		);

  reg [15:0] s1_inst;
  reg s1_predec;
  reg s1_postinc;
  reg [3:0] s1_addr_ah;

  assign INST = s1_inst;
`ifdef S0_INCLUDE_ROM
  wire [15:0] PM_OUT;
  rtavr_rom #(.SIZE(4096)) PM(.CLK(CLK) , .ADDRA(PC), .DOA(PM_OUT) , 
		 .WEB(WEB) , .OE(OE) , .ADDRB(ADDRB),
		 .DIB(DIB), .DOB(DOB) 
		);
`endif

//         LD.Rd.--X	1001:000d:dddd:1110
//         ST.--X.Rr	1001:001r:rrrr:1110
//         LD.Rd.--Y	1001:000d:dddd:1010
//         ST.--Y.Rr	1001:001r:rrrr:1010
//         LD.Rd.--Z	1001:000d:dddd:0010
//         ST.--Z.Rr	1001:001r:rrrr:0010
//         ELPM.Rd.Z    1001:000d:dddd:0110  << undefined instruction
  wire v_predec = ( PM_OUT[15:10] == 6'b100100 ) & ( PM_OUT[1:0] == 2'b10 );
//         LD.Rd.X++	1001:000d:dddd:1101
//         ST.X++.Rr	1001:001r:rrrr:1101
//         LD.Rd.Y++	1001:000d:dddd:1001
//         ST.Y++.Rr	1001:001r:rrrr:1001
//         LD.Rd.Z++	1001:000d:dddd:0001
//         ST.Z++.Rr	1001:001r:rrrr:0001
//         LPM.Rd.Z++   1001:000d:dddd:0101  << undefined instruction
  wire v_postinc = ( PM_OUT[15:10] == 6'b100100 ) & ( PM_OUT[1:0] == 2'b01 );
//       0xa,0xb: (26 - 16, 27 - 16)
//         LD.Rd.X	1001:000d:dddd:1100
//         ST.X.Rr	1001:001r:rrrr:1100
//         LD.Rd.X++	1001:000d:dddd:1101
//         ST.X++.Rr	1001:001r:rrrr:1101
//         LD.Rd.--X	1001:000d:dddd:1110
//         ST.--X.Rr	1001:001r:rrrr:1110
//         POP          1001:000d:dddd:1111  <<<
//         PUSH         1001:001d:dddd:1111  <<<
  wire v_xreg = ( PM_OUT[15:10] == 6'b100100 ) & ( PM_OUT[3:2] == 2'b11 )
                  & ( PM_OUT[1:0] != 2'b11 );
//       0xc,0xd:(28 - 16, 29 - 16)
//         LD.Rd.Y	1000:000d:dddd:1000
//         ST.Y.Rr	1000:001r:rrrr:1000
//         LD.Rd.Y++	1001:000d:dddd:1001
//         ST.Y++.Rr	1001:001r:rrrr:1001
//         LD.Rd.--Y	1001:000d:dddd:1010
//         ST.--Y.Rr	1001:001r:rrrr:1010
  wire v_yreg = ( PM_OUT[15:10] == 6'b100100 ) & ( PM_OUT[3:2] == 2'b10 )
		| ( PM_OUT[15:12] == 4'b1000 ) & ( PM_OUT[3:2] == 2'b10 );
//       0xe,0xf:(30 - 16, 31 - 16)
//         LD.Rd.Z	1000:000d:dddd:0000
//         ST.Z.Rr	1000:001r:rrrr:0000
//         LD.Rd.Z++	1001:000d:dddd:0001
//         ST.Z++.Rr	1001:001r:rrrr:0001
//         LD.Rd.--Z	1001:000d:dddd:0010
//         ST.--Z.Rr	1001:001r:rrrr:0010
  wire v_zreg = ( PM_OUT[15:10] == 6'b100100 ) & ( PM_OUT[3:2] == 2'b00 )
		| ( PM_OUT[15:12] == 4'b1000 ) & ( PM_OUT[3:2] == 2'b00 );
//       0xe,0xf:(30 - 16, 31 - 16)
  
//        0xe,0xf:(30 - 16, 31 - 16)
//          IJMP	1001:0100:0000:1001
//          ICALL	1001:0101:0000:1001
//          EIJMP       1001:0100:0001:1001  << undefined instruction
//          EICALL      1001:0101:0001:1001  << undefined instruction
  wire v_icall = ( PM_OUT[15:10] == 6'b100101 ) & PM_OUT[3] & PM_OUT[0];

`ifdef NO_STALL_IDX
// LOAD Rr (Rd -> Rr @s0)
// LD.Rd.Z         1000:000d:dddd:0000     # 
// LD.Rd.Y         1000:000d:dddd:1000     # 
// LD.Rd.Z++       1001:000d:dddd:0001     # 
// LD.Rd.--Z       1001:000d:dddd:0010     # 
// LD.Rd.Y++       1001:000d:dddd:1001     # 
// LD.Rd.--Y       1001:000d:dddd:1010     # 
// LD.Rd.X         1001:000d:dddd:1100     # 
// LD.Rd.X++       1001:000d:dddd:1101     # 
// LD.Rd.--X       1001:000d:dddd:1110     # 

// POP             1001:000d:dddd:1111     # 

// SYORE Rr
// ST.Z.Rr         1000:001r:rrrr:0000     # 
// ST.Y.Rr         1000:001r:rrrr:1000     # 
// ST.Z++.Rr       1001:001r:rrrr:0001     # 
// ST.--Z.Rr       1001:001r:rrrr:0010     # 
// ST.Y++.Rr       1001:001r:rrrr:1001     # 
// ST.--Y.Rr       1001:001r:rrrr:1010     # 
// ST.X.Rr         1001:001r:rrrr:1100     # 
// ST.X++.Rr       1001:001r:rrrr:1101     # 
// ST.--X.Rr       1001:001r:rrrr:1110     # 

// PUSH            1001:001d:dddd:1111     # 
//                 100X:00S-----------

// LPM.Rd.Z        1001:000d:dddd:0100     # undefined inst
// LPM.Rd.Z++      1001:000d:dddd:0101     # undefined inst
// ELPM.Rd.Z       1001:000d:dddd:0110     # undefined inst
// ELPM.Rd.Z++     1001:000d:dddd:0111     # undefined inst
// LDS(2word)      1001:000d:dddd:0000     # undefined inst
// STS(2word)      1001:001d:dddd:0000     # undefined inst

  // LD/ST + POP/PUSH
  wire v_ldst = (PM_OUT[15:13] == 3'b100) & (PM_OUT[11:10] == 2'b00);
`endif
  
assign WB_INST  =   v_predec | v_postinc;
`ifdef NO_STALL_IDX
`else
assign IDX_INST  =  v_xreg | v_yreg | v_zreg;
`endif
assign GPR_PREDEC  = s1_predec;
assign GPR_POSTINC = s1_postinc;
assign GPR_ADDRAL   = v_xreg ? 4'hA : v_yreg ? 4'hC    // (r26) (r28)
                    : v_zreg ? 4'hE : PM_OUT[7:4];     // (r30)

wire [3:0] v_addr_ah    = INST[7:4];
assign GPR_ADDRAH   = s1_addr_ah;

assign GPR_ADDRBL   = v_icall ? 4'hE : PM_OUT[7:4];       // (r30)

`ifdef NO_STALL_IDX
assign GPR_ADDRBH   = (v_icall | v_ldst & v_zreg) ? 4'hF  // (r31)
                      : (v_ldst & v_yreg) ? 4'hD          // (r29)
                      : (v_ldst & v_xreg) ? 4'hB          // (r27)
		      : PM_OUT[3:0];
`else
assign GPR_ADDRBH   = (v_icall) ? 4'hF : PM_OUT[3:0];     // (r31)
`endif


  always @(negedge CLK)
  begin
    if (RESET == 0)
      begin
	s1_inst    <= 0;
	s1_predec  <= 0;
	s1_postinc <= 0;
	s1_addr_ah <= 0;
      end
    else
      begin
	s1_inst <= PM_OUT;
	s1_predec  <= S0_VALID & v_predec;
	s1_postinc <= S0_VALID & v_postinc;
	s1_addr_ah <= v_addr_ah;
      end
  end

endmodule
